_smdFolder = ""
_actorxFolder = ""

_skipFolders = 0

-- SMD Export by LordAardvark
-- write <file> <string> - writes <string> to <file>
function write f s = ( 
	format "%" (s + "\n") to:f
	--print s
)

function ParentSpace parent child = (
	parentPos = parent.pos
	parentRot = parent.rotation
	childPos = child.pos
	childRot = child.rotation
	
	parentRotConj = quat -parentRot.x -parentRot.y -parentRot.z parentRot.w
	
	localPos = childPos - parentPos
	localRot = childRot * parentRotConj
	
	return #(localPos,quatToEuler localRot)
)
	
-- Rounds a float to a set number of decimal places
-- Note that SMD provides up to 6 digits of precision, and it requires decimal notation.
-- At values of E-5 and smaller, Maxscript automatically tostring's floats in scientific notation,
-- and to my knowledge, there is no way to force that into decimal.
-- So, for now, I stick to 4 digits of precision. It seems to be precise enough (and yields trivially smaller filesizes)
function roundFloat val dp = ( 
	-- http://www.scriptspot.com/forums/3ds-max/general-scripting/printing-out-float-values-to-a-few-decimal-points#comment-12190
	a = val * (10^dp)
	b = a as integer
	a = b as float / (10^dp)
	return a
)

-- Hash function
-- Using the multiplication method, as described here https://www.cs.hmc.edu/~geoff/classes/hmc.cs070.200101/homework10/hashfuncs.html
hash_A = (0.5 * ((sqrt 5) - 1))
hash_p = 9
hash_w = 32

-- Given a integer k, will hash it dependant on p. 2^p is how many buckets there are
function hash k = (
	s = floor(hash_A * 2^hash_w)
	x = k * s
	
	return 1 + (bit.shift x -(hash_w - hash_p)) -- bit.shift <a> <b> performs a left-shift of <b>. If <b> is negative, it is a right-shift.
																 -- Note that this function generates in the range of [0, 2^p - 1], but Maxscript is 1-indexed, so it needs to increase by 1
)

-- Homebrew hash function for strings. Probably isn't the best, which is why it relies on the above hash function
-- I'm not looking for the best security ever. I just need a semi-decent mapping.
-- By all means, I should use a smaller bucket size for materials. Source can only support 32 materials, and in the ~10 years I've been
-- doing Source model, I have never seen a model source file use in 48 materials, but memory is cheap, so I will use the same size-512 bucket
-- that the bone hash function uses.
function hashStr str = (
	k = 0
	for i=1 to str.count do (
		k = k + ((stricmp "z" str[i]) * i) -- So far as I know, Maxscript has no way to get the unicode value of a char. Doing a cmp to z is the closest I know of.
		if toUpper str[i] == str[i] then ( -- Maxscript has no case-sensitive cmp, though, so I have to treat uppercase characters separately.
			k = k + (32 * i)
		)
	)
	
	return hash k
)

-- Globals
SEP_SPACE = " "
SEP_TAB = "   " -- Tabs (\t) are fine, unless you have 5+ bones on a vert, then they cause exception violations. Let's just abstract this.

-- Caches
smdImEx_bones = #() -- A cache of Skin Bone Index -> Bone object ; note this cache needs to be rebuilt for every skinned object
smdImEx_boneIndexer = #() -- A cache of handle -> sequential index
smdImEx_boneByHandle = #() -- A cache of handle -> Bone object
smdImEx_boneIndex = 0 -- The current sequential index, incremented with use of smdImEx_boneIndexer via getIndexedBone

smdImEx_matIndexerFromDiffuse = #() -- A cache of diffuse name -> sequential index
smdImEx_matIndexerToStrHash = #() -- A cache of material sequential index -> diffuse name hash
smdImEx_matIndex = 0 -- The current material sequential index

smdImEx_cachedSkin = undefined -- The cached skin, for use with smdImEx and getBoneFromListIndex

-- Given a skin and an index from that skin, get the bone object. Cached if possible for speed.
-- Please don't use skins non-sequentially. This function has to rebuild the cache every time a different skin is used.
function getBoneFromListIndex oSkin ind = (
	-- Rebuild the cache, if necessary
	if oSkin != smdImEx_cachedSkin then (
		smdImEx_bones = #()
		smdImEx_cachedSkin = oSkin
	)
	
	-- See if it's already cached
	k = hash ind
	
	if smdImEx_bones[k] != undefined then ( -- If this is undefined, then it's definitely not cached.
		for o in smdImEx_bones[k] do (
			if o[1] == ind then (
				return o[2]
			)
		)
	) else ( -- First time something is put into this bucket, so we need to create the bucket.
		smdImEx_bones[k] = #()
	)
	
	-- Not cached, so let's fetch it.
	-- I wish there was a way to get the bone object from skinOps, but so far as I know, this is the only way.
	boneName = skinOps.GetBoneName oSkin ind 1
	
	 -- We're only considering bones and dummies
	allBones = for o in Geometry where (classof o == BoneGeometry or classof o == Dummy) collect o
	
	for b in allBones do (
		if b.name == boneName then (
			append smdImEx_bones[k] #(ind,b)
			return b
		)
	)
	
	return undefined
)

-- Given a handle, will return the sequential bone index associated with it, or generate a new sequential index if none is defined.
function getIndexedBone handle = (
	-- See if it's already cached
	k = hash handle
	
	if smdImEx_boneIndexer[k] != undefined then (
		for o in smdImEx_boneIndexer[k] do (
			if o[1] == handle then (
				return o[2]
			)
		)
	) else (
		smdImEx_boneIndexer[k] = #()
	)
	
	index = smdImEx_boneIndex
	
	append smdImEx_boneIndexer[k] #(handle,index)

	-- I _THINK_ maxscript has the a += 1 notation, but from what I recall, it's really hit and miss as to whether or not it works as expected...
	-- This is safer
	smdImEx_boneIndex = smdImEx_boneIndex + 1
	
	return index
)

-- Given a handle, returns the bone with that handle.
-- Note that while this function CAN update its cache, it should only be used after the bones have cached.
function getBoneByHandle handle = (
	k = hash handle
	
	if smdImEx_boneByHandle[k] != undefined then (
		for o in smdImEx_boneByHandle[k] do (
			if o[1] == handle then (
				return o[2]
			)
		)
	) else (
		smdImEx_boneByHandle[k] = #()
	)
	
	allBones = (for o in Geometry where (classof o == BoneGeometry or classof o == Dummy) collect o)
	for o in allBones do (
		if o.handle == handle then (
			append smdImEx_boneByHandle[k] #(o.handle,o)
			return o
		)
	)
	
	return undefined
)

smdImEx_matIndexerFromDiffuse = #() -- A cache of diffuse name -> sequential index
smdImEx_matIndexerToStrHash = #() -- A cache of material sequential index -> diffuse name hash
smdImEx_matIndex = 1 -- The current material sequential index

function getMatIndexFromDiffuse diffName = (
	-- See if it's cached
	k = hashStr diffName
	
	if smdImEx_matIndexerFromDiffuse[k] != undefined then (
		for o in smdImEx_matIndexerFromDiffuse[k] do (
			if o[1] == diffName then (
				return o[2]
			)
		)
	) else (
		smdImEx_matIndexerFromDiffuse[k] = #()
	)
	
	-- Otherwise, generate a new sequence index for this material, add it to the cache, and return it
	-- We also add a mapping of new index -> string hash, for reverse lookup
	newIndex = smdImEx_matIndex
	kInd = hash newIndex
	
	if smdImEx_matIndexerToStrHash[kInd] == undefined then (
		smdImEx_matIndexerToStrHash[kInd] = #()
	)
	
	append smdImEx_matIndexerFromDiffuse[k] #(diffName, newIndex)
	append smdImEx_matIndexerToStrHash[kInd] #(newIndex,k)
	
	smdImEx_matIndex = smdImEx_matIndex + 1
	
	return newIndex
)

function getDiffuseFromStrHash k matIndex = (
	if smdImEx_matIndexerFromDiffuse[k] != undefined then (
		for o in smdImEx_matIndexerFromDiffuse[k] do (
			if o[2] == matIndex then (
				return o[1]
			)
		)
	)

	return undefined -- If we have the material hash, then that implies we have the material index, which implies it should already be cached
)

function getDiffuseFromMatIndex matIndex = (
	-- See if it's cached
	k = hash matIndex
	
	if smdImEx_matIndexerToStrHash[k] != undefined then (
		for o in smdImEx_matIndexerToStrHash[k] do (
			if o[1] == matIndex then (
				return getDiffuseFromStrHash o[2] matIndex
			)
		)
	) 
	
	-- Otherwise, it's not here
	return undefined
)
	
-- A recursive function that will dump the "node" entry for a given object 'o'.
-- It will then recurse on all of o's children.
-- Note that it will only dump the 'o' object if it is a bone or a dummy.
-- This function also updates the smdImEx_boneByHandle cache as it dumps
function ExportChildren f o = (
	parentId = "-1" -- Parentless bones always have a parent ID of -1.
	if classof o.parent == BoneGeometry or classof o.parent == Dummy then ( parentId = ((getIndexedBone o.parent.handle) as string) )
	
	if classof o == BoneGeometry or classof o == Dummy then (
		write f (((getIndexedBone o.handle) as string) + " \"" + o.name + "\" "+parentId)
		
		append smdImEx_boneByHandle #(o.handle,o)
	)
	
	for c in o.children do (
		ExportChildren f c
	)
)

-- The bulk of the export function.
-- Given a filename and a list of objects to export, it will generate a SMD for all of the provided objects.
-- The objects must be EditableMesh.
-- Sorry EditablePoly users.
function ExportObjectsToSMD filename objs = (	
	-- Create the file
	f = createFile filename
	
	-- Clear all of the caches
	smdImEx_bones = #()
	smdImEx_boneIndexer = #() 
	smdImEx_boneByHandle = #()
	smdImEx_boneIndex = 0

	smdImEx_cachedSkin = undefined
	
	-- Generate internal data structures
	oMeshes = for o in objs where (classof o == Editable_Mesh) collect o
	oBones = for o in rootScene.world.children where (classof o == Dummy or classof o == BoneGeometry) collect o -- Collect all of the roots.
	
	print "OBONES: "
	print oBones
	
	-- Header
	write f "version 1"
	
	-- Nodes
	write f "nodes"
	for o in oBones do (
		ExportChildren f o
	)
	write f "end"
	
	-- Collect all of the bones we just dumped into ExportChildren
	oBones = for o in smdImEx_boneByHandle collect o[2]
		
	-- Skeleton
	write f "skeleton"
	write f "time 0"
	for o in oBones do (
		-- o.rotInParent does not always return the correct value
		-- this method, taken from  http://www.gamedev.net/topic/567949-maxscript-getting-vertex-weights/?view=findpost&p=4641243 , does
		-- o.posInParent seems to work fine, though.
		
		localTrans = o.transform
		if o.parent != undefined then (
			localTrans = localTrans * inverse(o.parent.transform)
		)
		
		if o.parent != undefined then (
			localSpace = ParentSpace o.parent o
			
			bonePos = localSpace[1]
			boneRot = quatToEuler localSpace[2]
		) else (
			bonePos = o.pos
			boneRot = quatToEuler o.transform.rotation
		)
		
		bonePos = localTrans.position
		boneRot = quatToEuler localTrans.rotation -- We need Euler angles. Thankfully, Maxscript provides a convenient function for this
		--boneScale = o.scale
		boneScale = o.scale
	
		print ("BoneScale (" + (o.name) + ") : " + (boneScale as string))
		
		write f (((getIndexedBone o.handle) as string) \
			+ SEP_TAB + ((roundFloat (boneScale.x * bonePos.x) 4) as string) \
			+ SEP_SPACE + ((roundFloat (boneScale.y * bonePos.y) 4) as string) \
			+ SEP_SPACE + ((roundFloat (boneScale.z * bonePos.z) 4) as string) \
			+ SEP_TAB + ((roundFloat (degToRad boneRot.x) 4) as string) \
			+ SEP_SPACE + ((roundFloat (degToRad boneRot.y) 4) as string) \
			+ SEP_SPACE + ((roundFloat (degToRad boneRot.z) 4) as string))		
	)
	write f "end"
	
	-- Triangles (oh boy)
	write f "triangles"
	for o in objs do (
		-- Print metadata, for my importer to be able to rebuild meshes as separate objects
		--write f ("//MDo " + obj.name)
		
		-- Select the Skin modifier, so as to avoid a stupid maxscript error
		select o
		max modify mode
		
		numFaces = getNumFaces o
		oMat = o.material
		oSkin = o.modifiers[#skin]
		
		for i=1 to numFaces do (
			matId = getFaceMatId o i
			
			-- Source models should only be using either Multi/Sub or Standard material. Handle each case separately.
			matName = "Undefined"
			if classof oMat == Multimaterial then (
				matName = filenameFromPath ((getSubMtl oMat matId).diffuseMap).filename
			)
			if classof oMat == Standardmaterial then (
				matName = filenameFromPath oMat.diffuseMap.filename
			)
			
			-- Get the vertices and vertex normals for this triangle
			faceVerts = getFace o i
			faceNormals = meshop.getFaceRNormals o i
			
			-- This is no correlation of mesh vert -> UV vert, but there is a correlation of mesh face -> UV face.
			-- So we have to get the UV verts from the UV face.
			uvFace = getTVFace o i
			uvVerts = #(uvFace.x,uvFace.y,uvFace.z) -- Collapse the UVW into an array, for easy indexing
			
			write f matName
			
			-- We are going to make the highest-weight bone be the root.
			-- That way, if the weights don't add up to one, they will go to the bone most likely to do the excess weight right:
			-- the dominant bone
			rootBone = 0 as string
			rootBoneWeight = 0
			
			for j=1 to 3 do ( -- Every triangle is explicitly 3 vertices, hence the hard-coded 3. Someone call me to maintain this script if geometry changes.
				-- Get the vertex id, position, and normal. Note we need to normalize the normal, which I find infinitely humorous.
				vertId = faceVerts[j]
				vertPos = meshop.getVert o vertId
				vertNorm = normalize (faceNormals[j])
				
				-- Again, there is no direct vert->uv vert correlation. So we do this round-about.
				-- Luckily, there is a face->uv face correlation
				-- And the order of the vertices in the triangle is the same as the order of the uv face.
				-- So this is a safe index to perform
				vertUvPos = getTVert o uvVerts[j]
				
				-- A vertex might have no bones weighing it, so default to this.
				boneWeightStr = ""
				
				-- If there's no skin on this object, we can't get the weights from it, so just skip all of this.
				if oSkin != undefined then (
					-- Get how many bones are weighing on this particular vertex.
					numBoneWeights = skinOps.GetVertexWeightCount oSkin vertId
					
					if numBoneWeights > 0 then (
						boneWeightStr = boneWeightStr + "    " + (numBoneWeights as string)
						
						for b=1 to numBoneWeights do (
							-- Get the "skin system index" of this bone, and its weight.
							sysBoneInd = skinOps.GetVertexWeightBoneID oSkin vertId b
							boneWeight = skinOps.GetVertexWeight oSkin vertId b
							
							-- Use the "skin system index" of the bone to get the actual bone, and its handle
							boneObject = getBoneFromListIndex oSkin sysBoneInd
							boneHandle = getIndexedBone boneObject.handle
							
							boneWeightStr = boneWeightStr \
								+ SEP_SPACE + (boneHandle as string) \
								+ SEP_SPACE + ((roundFloat boneWeight 4) as string)

							-- If this bone has more weight than our currently stored root, make it the root bone.
							if boneWeight > rootBoneWeight then (
								rootBone = boneHandle as string
								rootBoneWeight = boneWeight
							)
						)
					)
										
					write f (rootBone \
						+ SEP_TAB + ((roundFloat vertPos.x 4) as string) \
						+ SEP_SPACE + ((roundFloat vertPos.y 4) as string) \
						+ SEP_SPACE + ((roundFloat vertPos.z 4) as string) \
						+ SEP_TAB + ((roundFloat vertNorm.x 4) as string) \
						+ SEP_SPACE + ((roundFloat vertNorm.y 4) as string) \
						+ SEP_SPACE + ((roundFloat vertNorm.z 4) as string) \
						+ SEP_TAB + ((roundFloat vertUvPos.x 4) as string) \
						+ SEP_SPACE + ((roundFloat vertUvPos.y 4) as string) \ 
						+ boneWeightStr)
					)
			)
		)
	)
	write f "end"
	write f ""
	
	close f
)

-- Exports all of the meshes as SMD.
function ExportAllToSMD filename = (
	ExportObjectsToSMD filename (for o in Geometry where classof o == Editable_Mesh collect o)
)

function StitchMesh = (
	for o in Geometry do (
		if classof o == Editable_Mesh then (
			meshOp.weldVertsByThreshold o #{1..(o.numVerts)} 0.0001
		)
	)
)

function SmoothMesh = (
	for o in Geometry do (
		if classof o == Editable_Mesh then (
			meshop.autoSmooth o #{1..o.numfaces} 360
		)
	)
)

function EnsureSkin = (
	pos = point3 0 0 0
	endpos = point3 0 0 5
	bonerot = point3 0 0 1
	
	for o in Geometry do (
		if classof o == Editable_Mesh then (
			if o.modifiers[#skin] == undefined then (
				rootBone = bonesys.createBone pos endpos bonerot
				rootBone.name = o.name + "_root"
				
				s = Skin()
				addmodifier o s

				skinOps.addBone s rootBone 1
				
				for i=1 to o.numVerts do (
					skinOps.setVertexWeights s i 1 1
				)
			)
		)
	)
)

-- Batch port
rollout Batchport_Actorx_UI "ActorX Batch port" width:220 (	
	edittext inputFolderDisplay "ActorX Input:" text:"E:\Test\ActorXMini"
	button inputFolderBtn "Select ActorX Root Folder"	
	label spacer ""
	
	edittext outputFolderDisplay "SMD Output:" text:"E:\Test\ModelSrc"
	button outputFolderBtn "Select SMD Root Folder"
	
	label spacer2 ""
	
	button btnPort "Port Folder"
	
	on inputFolderBtn pressed do (
		destFolder = getSavepath caption:"Select the root folder to load PSKX from." initialDir:inputFolderDisplay.text
		
		if (destFolder == undefined) then return messageBox "ERROR: Destination-folder browser ended prematurely."

		_actorxFolder = destFolder
		
		inputFolderDisplay.text = destFolder
	)
	
	on outputFolderBtn pressed do (
		destFolder = getSavepath caption:"Select the root folder to save SMDs to." initialDir:outputFolderDisplay.text
		
		if (destFolder == undefined) then return messageBox "ERROR: Destination-folder browser ended prematurely."

		_smdFolder = destFolder
		
		outputFolderDisplay.text = destFolder
	)
	
	on btnPort pressed do (
		_actorxFolder = inputFolderDisplay.text
		_smdFolder = outputFolderDisplay.text
		
		curFolder = 0
		
		subfolders = getDirectories (_actorxFolder + "/*")
		
		for i=1 to subfolders.count do (
			if curFolder >= _skipFolders then (
				files = getFiles (subfolders[i] + "*")
				
				for j=1 to files.count do (
					folderSplit = filterString files[j] "\\" splitEmptyTokens:true
					fileName = folderSplit[folderSplit.count]
					fnameNoExtension = substring filename 1 (filename.count-5)
					if findstring filename "pskx" == undefined then (
						fnameNoExtension = substring filename 1 (filename.count-4)
					)
					exportName = _smdFolder + "/" + folderSplit[folderSplit.count - 1] + "/ref_" + fnameNoExtension + ".smd"

					-- Clear scene
					objs = #()
					for o in geometry do (
						if classof o == Editable_Mesh or classof o == BoneGeometry then (
							append objs o
						)
					)
					
					-- Because apparently I can't delete them straight out of geometry. Probably rearranges every time I delete, so things go out of bounds.
					for o in objs do (
						delete o
					)
				
					-- Import the ActorX
					try (
						ImportPskFile files[j]
						
						-- Fix up the ActorX
						StitchMesh()
						SmoothMesh()
						EnsureSkin()
						
						-- Export the SMD
						ExportAllToSMD exportName
					) catch (
						print ("Error porting \"" + exportName + "\".")
					)
				)
			)
			
			print ("Processed folder \"" + subfolders[i] + "\" (#" + (CurFolder as string) + ")")
			
			curFolder = curFolder + 1
		)
		
		print "Done!!"
	)
)

createDialog Batchport_ActorX_UI