---------------------------
-- Batch Import/ Export
-- By Gmod4ever
---------------------------
-- Not particularly useful by itself,
-- but when you need to import a large
-- amount of files, perform some operations to them
-- and then export them, this comes in handy
--
-- An example use-case is needing to batch-replace
-- the skeleton of bodygroups
--
-- Note that the storage of filenames does NOT persist
-- you must import and export in one sitting!
----------------------------

-- Dictionary definition
struct DictData (name,value)
struct Dictionary (
	_internal = #(),
	
	function Clear = (
		_internal = #()
	),
	
	function Count = (
		return _internal.Count
	),
	
	function Keys = (
		local k = #()
		
		for o in _internal do (
			if classof o == DictData then (
				append k o.name
			)
		)
		
		return k
	),
	
	function Values = (
		local k = #()
		
		for o in _internal do (
			if classof o == DictData then (
				append k o.value
			)
		)
	),
	
	function KeyValues = (
		local k = #()
		
		for o in _internal do (
			if classof o == DictData then (
				append k o
			)
		)
		
		return k
	),
	
	function IndexOf name = (
		for i = 1 to _internal.Count do (
			local o = _internal[i]
			
			if classof o == DictData then (
				if o.name == name then (
					return i
				)
			)
		)
		
		return -1
	),
	
	function Index ind = (
		return _internal[ind]
	),
	
	function Get name = (
		for o in _internal do (
			if classof o == DictData then (
				if o.name == name then (
					return o.value
				)
			)
		)
		
		return undefined
	),
	
	function Add name value = (
		local ind = IndexOf name
		
		if ind >= 0 then (
			_internal[ind] = value
		) else (
			local d = DictData()
			d.name = name
			d.value = value
			
			append _internal d
		)
	)
)

-- Globals
global BIEDict = Dictionary()

-- Helper functions
function getFilename fPath = (
	fName = filenameFromPath fPath
	fLength = fName.count
	
	fName = substring fName 1 (fLength - 4)
	
	return fName
)

function hideAllBut emesh = (
	for o in Geometry do (
		if classof o == Editable_Mesh then (
			if o != emesh then (
				o.ishidden = true
			)
		)
	)
)

function unhideAll = (
	for o in Geometry do (
		if classof o == Editable_mesh then (
			o.ishidden = false
		)
	)
)

-- Core
function doImport fileName = (
	-- First, actually import the file
	ImportFile fileName
	
	for o in selection do (
		print o
	)
	
	-- Then, grab the mesh
	local emesh = undefined
	
	local potentials = getnodebyname "SMDImport" all:true
	
	for o in potentials do (
		if classof o == Editable_Mesh then (
			emesh = o
			break
		)
	)
	
	if emesh != undefined then (
		-- Rename the mesh
		emesh.name = getFilename fileName
	
		-- Delete any other "SMDImports", to avoid any collisions
		local potentials = getnodebyname "SMDImport" all:true
		
		for o in potentials do (
			delete o
		)
	
		-- Add a fileName -> mesh reference
		BIEDict.Add fileName emesh
	)
)

function doExport fileName = (
	if BIEDict.IndexOf fileName >= 0 then (
		-- Get our target mesh
		local emesh = BIEDict.Get fileName
		
		-- Hide all the editable meshes that are not our mesh
		hideAllBut emesh
		
		-- Export this mesh to a separate "out" folder
		local cleanedName = trimLeft (trimRight (filenameFromPath fileName) " \n\t\"") " \n\t\""

		local outPath = (getFilenamePath fileName) + "out\\"
		local outFile = outPath + cleanedName
		
		MakeDir outPath all:true
		
		ExportFile outFile
		
		-- Then unhide everything
		unhideAll()
	)
)

-- UI
rollout BatchImportExport "Batch Import/Export" (	
	
	button btnImport "Import"
	button btnExport "Export"
	
	on btnImport pressed do (
		-- Reset the dictionary
		BIEDict.Clear()
		
		-- Prompt the user for files to import
		fileDialog = dotNetObject "System.Windows.Forms.OpenFileDialog"
		fileDialog.title = "Select Files to Import"
		fileDialog.Multiselect = true
		
		onClose = fileDialog.showDialog()
		
		-- If the selection was not cancelled...
		if onClose.Equals onClose.OK then (
			files = fileDialog.fileNames
			
			-- Import the files
			imported = 0
			for file in files do (
				doImport file
				imported = imported + 1
			)
			
			MessageBox ("Imported " + (imported as String) + " objects.") beep:false
		)
	)
	
	on btnExport pressed do (
		-- Loop through the dictionary and export everything
		keys = BIEDict.Keys()
		
		exported = 0
		for fileName in keys do (
			doExport fileName
			exported = exported + 1
		)
		
		MessageBox ("Exported " + (exported as String) + " objects.") beep:false
	)
)

createDialog BatchImportExport