-- Initialize
-- These are all persistent-global, so they will save in .max files
-- The checks are so that they will initialize in a new file, but not override for files that already have them
persistent global ExportGroups
	
if ExportGroups == undefined then (
	ExportGroups = #()
)

persistent global AlwaysExport
	
if AlwaysExport == undefined then (
	AlwaysExport = #()
)

persistent global ExportPath
	
if ExportPath == undefined then (
	ExportPath = "C:/"
)

GroupsBox = undefined

curVisibilityState = #()

function GetObjsByNames names = (
	local rarray = #()
	
	for o in Geometry do (
		if findItem names o.name != 0 then (
			append rarray o
		)
	)
	
	return rarray
)

-- Exports all of the meshes as SMD.
function ExportAllToSMD filename = (
	ExportObjectsToSMD filename (for o in Geometry where classof o == Editable_Mesh collect o)
)

function pushVisibility = (
	curVisibilityState = #()
	for o in Geometry do (
		append curVisibilityState #(o,o.ishidden)
	)
)

function popVisibility = (
	for o in curVisibilityState do (
		o[1].ishidden = o[2]
	)
)

function getFilename fPath = (
	fName = filenameFromPath fPath
	fLength = fName.count
	
	fName = substring fName 1 (fLength - 4)
	
	return fName
)

-- Update the always-export to be everything that isn't an editable mesh.
-- Probably not the best way to do it, but fuck it.
function UpdateAlwaysExport_Generic = (
	AlwaysExport = #()
	for o in Objects do (
		if classof o != Editable_Mesh and classof o != Editable_Poly and classof o != PolyMeshObject then (
			append AlwaysExport o
		)
	)
)

-- Get the ID for a given export-group name, or -1 if it doesn't exist
function GetExportGroup GroupName = (
	for i = 1 to ExportGroups.Count do (
		if ExportGroups[i][1] == GroupName then (
			return i
		)
	)
	
	return -1
)

-- Create a new export group with the provided name for what's selected
-- Or update an already-existing group to include the newly selected objects
function GE_CreateExportGroup GroupName bNotify:true = (
	grpId = GetExportGroup GroupName
	
	grp = undefined
	
	if grpId == -1 then (
		grp = #(GroupName)
		grpId = ExportGroups.Count + 1
	) else (
		grp = ExportGroups[grpId]
	)
	
	
	for o in selection do (
		appendifunique grp o.name
	)
	
	cnt = grp.Count - 1
	
	if bNotify then (
		messageBox ("Group " + GroupName + " has " + (cnt as string) + " elements") beep:false
	)
	
	ExportGroups[grpId] = grp
)

-- Update the list-box with all existing groups
function UpdateExistingGroups = (
	displayItems = #()
	
	for grp in ExportGroups do (
		grpName = grp[1]
		
		append displayItems grpName
	)
	
	print displayItems
	
	GroupsBox.items = displayItems
)

function DoGroupExport ids bOverwrite = (
	_exportGroups = #()
	
	for id in ids do (
		grp = ExportGroups[id]
		
		grpObjs = #()
		grpName = grp[1]
		
		for i=2 to grp.Count do (
			append grpObjs grp[i]
		)
		
		grpObjs = GetObjsBynames grp
		
		selectionGrp = #()
		join selectionGrp AlwaysExport
		join selectionGrp grpObjs
		
		exportName = ExportPath + "/" + grpName + ".smd"
		
		if bOverwrite == true then (
			deleteFile exportName
		)
		
		_exportGroup = createExportGroup exportName selectionGrp

		append _exportGroups _exportGroup
	)
	
	ExportSMDSilentBatch _exportGroups bKeepBones:false
)

-- UI
rollout GroupExporter_UI_CreateGroup "Create Group" width:220 (
	edittext txtGroupName "Group Name: "
	button btnCreate "Create"
	
	on btnCreate pressed do (
		GE_CreateExportGroup txtGroupName.text
		
		UpdateExistingGroups()
		
		DestroyDialog GroupExporter_UI_CreateGroup
	)
)

rollout GroupExporter_UI_ExportGroups "Export Groups" width:220 (
	edittext txtExportPath "Export Path:"
	button btnExportPath "Set Path"
	multilistbox exportable "Select Groups to Export"
	button btnAll "Select All"
	button btnNone "Select None"
	button btnExport "Export Selected Groups"
	checkbox chkOverwrite "Overwrite Existing Models?"
	
	on GroupExporter_UI_ExportGroups open do (
		exportable.items = GroupsBox.items
		txtExportPath.text = ExportPath
	)
	
	on btnExportPath pressed do (
		getExportPath = getSavepath caption:"Select the folder to export SMDs to." initialDir:txtExportPath.text
		
		if (getExportPath == undefined) then return messageBox "ERROR: Destination-folder browser ended prematurely."
		
		ExportPath = getExportPath
		txtExportPath.text = getExportPath
	)
	
	on btnAll pressed do (
		selectItems = # ()
		for i = 1 to exportable.items.Count do (
			append selectItems i
		)
		
		exportable.selection = selectItems
	)
	
	on btnNone pressed do (
		exportable.selection = #{}
	)
	
	on btnExport pressed do (
		DoGroupExport exportable.selection chkOverwrite.checked
		
		messagebox ("Exported " + (exportable.selection.count as string) + " groups.") beep:false
	)
)

rollout GroupExporter_UI "Group Exporter" width:220 (
	listbox existingGroups "Existing Export Groups"
	button btnUpdate "Update Always-Export"
	button btnCreate "Create Group"
	button btnCreateFromSelection "Create Groups From Selection"
	button btnSelect "Select Group"
	button btnAdd "Add to Selected Group"
	button btnHide "Hide All But Selected Group"
	button btnDelete "Delete Group"
	button btnClear "Clear Groups"
	button btnExport "Export Groups"
	
	on GroupExporter_UI open do (
		GroupsBox = existingGroups
		
		UpdateExistingGroups()
	)
	
	on btnUpdate pressed do (
		UpdateAlwaysExport_Generic()
		
		messageBox ( "There are " + (AlwaysExport.Count as string) + " elements that will always be exported." ) beep:false
	)
	
	on btnCreate pressed do (
		createDialog GroupExporter_UI_CreateGroup
	)
	
	on btnCreateFromSelection pressed do (
		selObjs = #()
		for o in selection do (
			append selObjs o
		)
		
		for o in selObjs do (
			clearSelection
			
			select o
			
			gname = o.name
			gname = substituteString o.name " " "_"
			gname = toLower gname
			
			GE_CreateExportGroup gname bNotify:false
		)
		
		UpdateExistingGroups()
	)
		
		
	on btnSelect pressed do (
		selInd = GroupsBox.selection
		
		grp = ExportGroups[selInd]
		
		clearSelection
		selGrp = #()
		for i=2 to grp.Count do (
			append selGrp grp[i]
		)
		
		selGrpObjs = GetObjsByNames selGrp
		
		select selGrpObjs
	)
	
	on btnAdd pressed do (
		selInd = GroupsBox.selection
		
		grp = ExportGroups[selInd]

		GE_CreateExportGroup grp[1]
	)
	
	on btnHide pressed do (
		selInd = GroupsBox.selection
		
		grp = ExportGroups[selInd]
		
		selGrp = #()
		for i=2 to grp.Count do (
			append selGrp grp[i]
		)
		
		selGrpObjs = GetObjsByNames selGrp
		hideAllBut selGrpObjs
	)
	
	on btnDelete pressed do (
		selInd = GroupsBox.selection
		
		deleteitem ExportGroups selInd
		
		UpdateExistingGroups()
	)
	
	on btnClear pressed do (
		ExportGroups = #()
		
		UpdateExistingGroups()
	)
	
	on btnExport pressed do (
		createDialog GroupExporter_UI_ExportGroups
	)
)

createDialog GroupExporter_UI
