
VtaToMeshMap = #()
VtaFrames = #()

VtaNames = #()
VtaSplits = #()

MeshCacheChunks = #()
MeshCacheChunkSize = 2

_eps = 0.001
fn IsSameVert vertA vertB = (
	return ((abs (vertA[1] - vertB[1])) <= _eps) and ((abs (vertA[2] - vertB[2])) <= _eps) and ((abs (vertA[3] - vertB[3])) <= _eps)
)

fn FlattenIndex x y z maxx maxy = (
	return (x + maxy * (y + maxx * z)) + 1
)

fn GetCacheBounds oMesh = (
	local xminbound = floor(oMesh.min.x / MeshCacheChunkSize) * MeshCacheChunkSize
	local xmaxbound = ceil(oMesh.max.x / MeshCacheChunkSize) * MeshCacheChunkSize
	
	local yminbound = floor(oMesh.min.y / MeshCacheChunkSize) * MeshCacheChunkSize
	local ymaxbound = ceil(oMesh.max.y / MeshCacheChunkSize) * MeshCacheChunkSize

	local zminbound = floor(oMesh.min.z / MeshCacheChunkSize) * MeshCacheChunkSize
	local zmaxbound = ceil(oMesh.max.z / MeshCacheChunkSize) * MeshCacheChunkSize
		
	return #( (point3 xminbound yminbound zminbound),
		(point3 xmaxbound ymaxbound zmaxbound),
		(point3 ((xmaxbound - xminbound) / MeshCacheChunkSize) ((ymaxbound - yminbound) / MeshCacheChunkSize) ((zmaxbound - zminbound) / MeshCacheChunkSize))
	)
)

fn GetCacheChunkId oMesh pVert = (
	local bounds = GetCacheBounds oMesh
	
	local xcoord = ((floor(pVert.x / MeshCacheChunkSize) * MeshCacheChunkSize) - bounds[1].x) / MeshCacheChunkSize
	local ycoord = ((floor(pVert.y / MeshCacheChunkSize) * MeshCacheChunkSize) - bounds[1].y) / MeshCacheChunkSize
	local zcoord = ((floor(pVert.z / MeshCacheChunkSize) * MeshCacheChunkSize) - bounds[1].z) / MeshCacheChunkSize
	
	return (FlattenIndex xcoord ycoord zcoord bounds[3].x bounds[3].y)
)

fn GenerateMeshCache oMesh = (
	print ("Generating cache for mesh \"" + oMesh.name + "\" ...")
	
	MeshCacheChunks = #()
	
	local numVerts = getnumverts oMesh
	local bDebugChunks = false
	
	for i=1 to numVerts do (
		local vert = getvert oMesh i
		
		local chunkId = GetCacheChunkId oMesh vert
		
		if MeshCacheChunks[chunkId] == undefined then (
			MeshCacheChunks[chunkId] = #()
		)
		
		append MeshCacheChunks[chunkId] #(i,vert)
	)
	
	print "Completed caching."
	
	if bDebugChunks then (
		for i=1 to MeshCacheChunks.count do (
			if MeshCacheChunks[i] == undefined then (
				--print ("Chunk " + (i as string) + " is empty.")
			) else (
				print ("Chunk " + (i as string) + " contains " + (MeshCacheChunks[i].count as string) + " verts.")
			)
		)
	)
	
)

fn ResetDataStructs = (
	VtaToMeshMap = #()
	VtaFrames = #()
	VtaNames = #()
	VtaSplits = #()
	
	MeshCacheChunks = #()
)

fn ReadVTA oMesh fName = (
	-- Note: Should probably reset the data structs first, but it's not strictly speaking necessary.
	
	f = openFile fName
	
	local meshCache = #()
	
	local bInFrames = false
	local iCurFrame = -1
	
	local numVerts = getnumverts oMesh
	
	local debugPulseRate = 3000 -- ms
	local debugDoPulse = true
	
	local updateTimeStart = timestamp()
	
	local bBuildMap = true
	if (VtaToMeshMap != undefined) and (VtaToMeshMap.count > 0) then (
		bBuildMap = false
	)
	local vtaVerts = 0
	
	if (MeshCacheChunks.count < 1) then (
		print "Precaching mesh ..."
		
		GenerateMeshCache oMesh
	)
	
	print ("Reading \"" + fName + "\" ...")
	
	while not (eof f) do (
		l = readLine f
		
		-- Keep 3dsmax from timing out.
		--if (timestamp() - updateTimeStart) >= debugPulseRate then (
		--	updateTimeStart = timestamp()
		--	print "Pulse"
		--	windows.processPostedMessages()	
		--)	
					
		-- Just chew until we get into a vertex frame
		if ((not bInFrames) and ((findString l "vertexanimation") != undefined)) then (
			print ("Entering frames on line \"" + l + "\".")
		
			windows.processPostedMessages()	
		
			bInFrames = true
		
			continue
		)
		
		-- If we're in the frames, and we find an end, then we're done.
		if (bInFrames) and ((findString l "end") != undefined) then (
			print ("Exiting frames on line \"" + l + "\".")
		
			windows.processPostedMessages()	
		
			exit
		)
		
		-- Otherwise, if we're in the frames, then we're reading either vertices or frames.
		if (bInFrames) then (
			
			local split = filterString l " "
			
			-- "frame d" -> set iCurFrame to d
			if (split[1] == "time") then (
				iCurFrame = (split[2] as integer) + 1
				
				VtaFrames[iCurFrame] = #() -- Each frame gets its own array objects, (vtaVertId, position)
				VtaNames[iCurFrame] = ("flex " + (iCurFrame as string))
				VtaSplits[iCurFrame] = false
					
				print ("Beginning frame " + (iCurFrame as string) + " on line \"" + l + "\".")
				
				windows.processPostedMessages()	
				
			) else ( -- Otherwise, this is an actual vertex
				local vertId = (split[1] as integer) + 1
				
				local x = split[2] as float
				local y = split[3] as float
				local z = split[4] as float
				
				local vert = point3 x y z
								
				-- Keep 3dsmax from timing out.
				if debugDoPulse and (timestamp() - updateTimeStart) >= debugPulseRate then (
					updateTimeStart = timestamp()
					print ("Pulse; Storing vert [" + (vertId as string) + "] = " + (vert as string) +" for frame " + (iCurFrame as string) + " on line \"" + l + "\".")
					windows.processPostedMessages()	
				)

				local frameObj = #(vertId, vert)
				
				append vtaFrames[iCurFrame] frameObj
				
				-- We only build the vertex map on the first frame.
				if iCurFrame == 1 and bBuildMap then (
					if (vertId <= numverts) then (
						vtaVerts = vtaVerts + 1
						
						local chunkId = GetCacheChunkId oMesh vert
						
						if MeshCacheChunks[chunkId] != undefined then (
							-- And now we go through all of the verts in the chunk
							for o in MeshCacheChunks[chunkId] do (
								local meshVertId = o[1]
								local meshVert = o[2]
								
								if IsSameVert vert meshVert then (
									if VtaToMeshMap[vertid] == undefined then (
										VtaToMeshMap[vertid] = #()
									)
									
									append VtaToMeshMap[vertid] meshVertId
								)
							)
							
						) else (
							print "WARNING!! Attempted to index empty chunk for vertex " + (vert as string) + " ( " + (vertId as string) + " )! Manually loading..."
							
							for i=1 to numVerts do (
								--local oVert = getvert oMesh i
								local oVert = meshCache[i]
								
								if IsSameVert vert oVert then (
									if VtaToMeshMap[vertid] == undefined then (
										VtaTomeshMap[vertid] = #()
									)
									
									append VtaToMeshMap[vertid] i -- There might be duplicate verts on the base mesh, so we have to map an array. :x
								)
							)
						)
					)
				)
			)
		)
	)
	
	close f
	
	if bBuildMap then (
		-- Because I got lost in parenthetical Hell, and caused this fucking string to break on me.
		-- So I'm going to play it First Grade and just have broken down into explicit steps.
		
		local frac = (vtaVerts as float) / (VtaToMeshMap.count as float)
		local diff = 1 - frac
		local percent = diff * 100.0
		
		print ( (VtaToMeshMap.count as string) + " of " + (vtaVerts as string) + " vta vertices mapped. " + (percent as string) + "% mapping.")
	)
	
	print "Done."
)

fn ShouldIgnore str = (
	local ignorestrs = #("//","defaultflex")
	
	for istr in ignorestrs do (
		if str == istr then (
			return true
		)
	)
	
	false
)

fn ReadQCI oMesh fName = (
	f = openFile fName
	
	print ("Reading QC/I \"" + fName + "\".")
	
	local bInFlexFile = false
	
	while not (eof f) do (
		l = readline f
		
		-- Flexfile? Load the VTA, and start loading the flex names & splits.
		if (findString l "flexfile") != undefined then (
			bInFlexFile = true
			
			local split = filterString l "\t "
			
			local vtaName = substituteString split[2] "\"" ""

			local folderPath = getFilenamePath fName
			
			local vtaPath = folderPath + vtaName
			
			ReadVTA oMesh vtaPath
			
			continue
		)
		
		-- Found a closing '}'? Then stop reading flex names & splits.
		if bInFlexFile and (findString l "}") != undefined then (
			bInFlexFile = false
			
			continue
		)
		
		-- Neither of the above, and we're not reading flex data? Then just skip this line. We don't care about it.
		if not bInFlexFile then (
			continue
		)
		
		-- Flexpair? Load the name and set its split to true.
		if (findString l "flexpair") != undefined then (
			local split = filterString l "\t "
			
			local flexName = substituteString split[2] "\"" ""
			local iFrame = (split[5] as integer) + 1
			
			VtaNames[iFrame] = flexName
			VtaSplits[iFrame] = true
			
			continue
		)
		
		-- Flex? Load the name and set its split to false.
		if (findString l "flex") != undefined then (
			local split = filterString l "\t "
			
			if not ShouldIgnore split[1] then ( -- Ignore defaultflex, comments, etc
				local flexname = substituteString split[2] "\"" ""
				
				local iFrame = (split[4] as integer) + 1
				
				VtaNames[iFrame] = flexName
				VtaSplits[iFrame] = false
				
				continue
			)
		)
	)
	
	print "Done."
)

-- iSide:
	-- 0	=	Left
	-- 1	= 	Right
	-- 2	=	Mono

fn LoadVtaFrame oMesh iFrame iSide:2 = (
	local bitVertIds = #{}
	local arrVertIds = #()
	local vertPoses = #()
	
	for o in VtaFrames[iFrame] do (
		local vtaVertId = o[1]
		local vert = o[2]
		
		local vertIds = VtaToMeshMap[vtaVertId]
		
		if vertIds != undefined then (
			if (iSide == 2) or (iSide == 0 and vert[1] >= 0) or (iSide == 1 and vert[1] < 0) then (
				for vertId in vertIds do (
					
					-- This "more efficient way" caused some morphs to distort. So we'll stick to the old, inefficient way. It's fast enough.

					--if (findItem bitVertIds vertId) < 1 then (
					--	append bitVertIds vertId
					--	append vertPoses vert
					--)
					
					meshop.setVert oMesh vertId vert
				)
			)
		)
	)
	
	--meshop.SetVert oMesh bitVertIds vertPoses
	
	--redrawViews() 
)

fn GenerateMorphFromFrame oMorpher oMasterShape iFrame iChannel = (
	if not VtaSplits[iFrame] then ( -- Mono morph
		oShape = undefined
		
		maxops.cloneNodes oMasterShape cloneType:#copy newNodes:&oShape
		
		oShape.name = VtaNames[iFrame]
		
		LoadVtaFrame oShape[1] iFrame
		
		WM3_MC_BuildFromNode oMorpher iChannel oShape[1]
		
		delete oShape
		
		print ("Generated mono morph \"" + VtaNames[iFrame] + "\".")
		windows.processPostedMessages()	
		
		return 1
	) else ( -- Stereo morph
		-- Left
		oShape = undefined
		maxops.cloneNodes oMasterShape cloneType:#copy newNodes:&oShape
		oShape.name = "left_" + VtaNames[iFrame]
		
		LoadVtaFrame oShape[1] iFrame iSide:0
		
		WM3_MC_BuildFromNode oMorpher iChannel oShape[1]
		
		delete oShape
		
		print ("Generated stereo left morph \"" + VtaNames[iFrame] + "\".")
		windows.processPostedMessages()	
		
		-- Right
		maxops.cloneNodes oMasterShape cloneType:#copy newNodes:&oShape
		oShape.name = "right_" + VtaNames[iFrame]

		LoadVtaFrame oShape[1] iFrame iSide:1
		
		WM3_MC_BuildFromNode oMorpher (iChannel+1) oShape[1]
		
		delete oShape
		
		print ("Generated stereo right morph \"" + VtaNames[iFrame] + "\".")
		windows.processPostedMessages()	
		
		return 2
	)
)

fn GetFirstFreeMorphChannel oMorpher = (
	for i=1 to 100 do (
		if not (WM3_MC_HasData oMorpher i) then (
			return i
		)
	)
	
	return undefined
)

fn GenerateMorpher oMesh lstMorphs:undefined = (
	local curMorphChannel = 1
	
	if oMesh.modifiers[#morpher] == undefined then (
		-- Create the Morpher
		oMorpher = Morpher()
		addModifier oMesh oMorpher
	) else (
		oMorpher = oMesh.modifiers[#morpher]
		curMorphChannel = GetFirstFreeMorphChannel oMorpher
	)
	
	-- Create the master shape, with a collapsed stack. All frames will be copied from it, to reduce the time it takes to copy morphers.
	masterShape = snapshot oMesh
	
	-- And then make a duplicate of the mesh for each frame, loading its values, and then storing it into the morpher.
	if lstMorphs == undefined then (
		local numFlexes = VtaFrames.count
		for i=2 to numFlexes do (
			if VtaFrames[i].count > 0 then ( -- There might be empty frames. No reason to add them to the Morpher. So only do this to frames with data.
				local newFrames = GenerateMorphFromFrame oMorpher masterShape i curMorphChannel
				
				curMorphChannel = curMorphChannel + newFrames
				
				if curMorphChannel > 100 then (
					messagebox "Ran out of morph channels!" beep:true
					exit
				)
			)
		)
	) else (
		for i in lstMorphs do (
			if VtaFrames[i].count > 0 then ( -- There might be empty frames. No reason to add them to the Morpher. So only do this to frames with data.
				local newFrames = GenerateMorphFromFrame oMorpher masterShape i curMorphChannel
				
				curMorphChannel = curMorphChannel + newFrames
				
				if curMorphChannel > 100 then (
					messagebox "Ran out of morph channels!" beep:true
					exit
				)
			)
		)			
	)
	
	delete masterShape
)

rollout VtaImporter_UI "QC/I + VTA Importer" width: 240 (
	multilistbox lstToBeLoaded "Flexes To Be Loaded:" height:8
	button btnDeleteFlex "Delete Selected" height: 16 pos:[124,4]
	pickbutton pbtnPickObject "Pick Mesh to Load Onto" width:208 pos:[16,140]
	button btnLoadQci "Load QCI" enabled:false width:208
	button btnGenerateMorphs "Generate Morphs" enabled:false width:208
	button btnReset "Reset" width:208
	
	on pbtnPickObject picked obj do (
		pbtnPickObject.text = obj.name
		pbtnPickObject.enabled = false
		
		btnLoadQci.enabled = true
	)
	
	on btnLoadQci pressed do (
		local qciName = getOpenFilename caption:"Choose a QC or QCI to load flexes from." types:"QC+QCI|*.qc*|QC|*.qc|QCI|*.qci|Any file|*.*"
		
		btnLoadQci.enabled = false
		
		ReadQCI pbtnPickObject.object qciName
		
		local displayItems = #()
		
		for i=2 to VtaFrames.count do (
			if VtaFrames[i].count > 0 then (
				append displayItems VtaNames[i]
			)
		)
		
		lstToBeLoaded.items = displayItems
		
		btnGenerateMorphs.enabled = true
	)
	
	on btnDeleteFlex pressed do (
		-- First get the string names of what will be deleted, since deletion is done by index and the box automatically updates itself every time
		local deleteByName = #()
		for o in lstToBeLoaded.selection do (
			append deleteByName lstToBeLoaded.items[o]
		)
		
		-- Then, for each item to be deleted, go through all the items until we find a match
		local items = lstToBeLoaded.items
		
		for sName in deleteByName do (
			for i=1 to items.count do (
				if items[i] == sName then (
					deleteItem items i
				)
			)
		)
		
		-- And then finally update the listbox
		lstToBeLoaded.items = items
	)
	
	on btnGenerateMorphs pressed do (
		local flexFrames = #()
		
		-- Match each frame by name
		local items = lstToBeLoaded.items
		
		for o in items do (
			for i=1 to VtaNames.count do (
				if VtaNames[i] == o then (
					append flexFrames i
				)
			)
		)
		
		btnGenerateMorphs.enabled = false
		
		-- And then feed that list of frames to the Morpher Generator
		GenerateMorpher pbtnPickObject.object lstMorphs:flexFrames
	)
	
	on btnReset pressed do (
		ResetDataStructs()
		
		lstToBeLoaded.items = #()
		pbtnPickObject.text = "Pick Mesh to Load Onto"
		
		pbtnPickObject.enabled = true
		btnLoadQci.enabled = false
		btnGenerateMorphs.enabled = false
	)
)

-- Testing
fn TestLoad = (
	ReadQCI geometry[1] "E:\Tools\Decompiled\Decompiled\Mine\Daz3D\DOA\_FantasyBase\Tina\flex.qci"
	--ReadVTA geometry[1] "E:\Tools\Decompiled\Decompiled\Mine\Daz3D\DOA\_FantasyBase\Tina\flex.vta"
)

createDialog VtaImporter_UI